EntropicGWdistLattice<-function(x.weight, y.weight, epsilon = 10^-3, x.size = c(1,1), y.size =c(1,1) ,options=list()){
  #' @export
  #' @title Computing the entropic Gromov-Wasserstein distance between two lattices
  #' @description The function \code{EntropicGWdistLattice} computes the
  #' the entropic Gromov-Wasserstein distance between 2 regular lattices X and Y with mass in the middle of each lattice box via
  #' projected gradient descent (cf. “Gromov-Wasserstein Averaging of Kernel and Distance Matrices.” by Peyré,
  #' Gabriel, Marco Cuturi, and Justin Solomon In ICML 2016, 2016.).
  #'
  #'
  #' @param  x.weight: The weigths of points in X as matrix.
  #' @param  y.weight: The weigths of points in Y as matrix.
  #' @param  epsilon: A numeric parameter of the entropic Gromov-Wasserstein distance.
  #' @param x_size: A 2-dim. vector containing the size of the lattice X. Here size(1) is the length of the grid
  #' in "x-direction" and size(2) in "y-direction".
  #' @param y_size: A 2-dim. vector containing the size of the lattice Y (cf. explanation for X).
  #' @param  options: A named list containing several options concerning the methods of computation.
  #' Possible options are:
  #'
  #' \describe{
  #' \item{"niter"}{The max. number of projected gradient descent steps ({default = 100}).}
  #' \item{"niter_sinkhorn"}{The max. number of steps in order to calculate a projection ({default = 400}).}
  #' \item{"gw_loss"}{A string marking the choice of the used loss. Possibilities are the well known L2 loss "L2" (default) and the Kullback-Leibler divergence "KL".}
  #' \item{"tol_gw"}{A numeric value in order to stop if the progress in the projection steps is smaller than this value (default = 10^-6).}
  #' \item{"safe"}{An option in order to increase the stability of the calculations ({default = 0}, choose 1 to enable).}
  #' \item{"gamma_init"}{A possible start coupling (as matrix) in order to speed up the calculations.}
  #' \item{"tol_sinkhorn"}{A numeric value in order to stop if the progress in the calculation of a projection steps is smaller than this value  (defalut =10^-10).}
  #' \item{"progress"}{A binary option to enable/disable a progressbar (default = 0, choose 1 to enable).}
  #' \item{"control"}{A binary option to control the progress of the calculations (default = 1). Setting this to 0 will speed up the calculations, but prevent them from being cancelled via the R console.}
  #' }
  #' @return A List containing a vector with the values of the projection steps while calculating the entopic Gromov-Wasserstein distance  between the 2  lattices X and Y ($distance) and the final coupling ($gamma).
  #' @examples
  #' An example how to call the described function:
  #'
  #' set.seed(13)
  #'
  #' options = list("niter"=50,"niter_sinkhorn"= 200, tol_gw = 1e-5, "tol_sinkhorn"= 1e-5, "gw_loss"= "L2")
  #' epsilon = 10^(-1)
  #' x.weight=matrix(rexp(150), ncol = 10, nrow = 15)
  #' y.weight=matrix((rnorm(150))^2, ncol = 3, nrow = 6)
  #'
  #' results = EntropicGWdistLattice(x.weight, y.weight, epsilon,c(10,1),c(13,7), options)

  if(!(is.list(options)&is.matrix(x.weight)&is.matrix(y.weight))){
    stop("Wrong input types!")
  }
  # finding the dimensions of the corresponding lattices
  x.dim <- dim(x.weight)
  y.dim <- dim(y.weight)
  # norming of the matrices and vectorize them in order
  # to call the Rcpp-function PerformGWSinkhorn
  x.weight <- as.vector(x.weight/sum(x.weight))
  y.weight <- as.vector(y.weight/sum(y.weight))

  x.distmat=distance.matrix.grid(x.dim[1],0,x.size[1],x.dim[2],0,x.size[2])$dist
  y.distmat=distance.matrix.grid(y.dim[1],0,y.size[1],y.dim[2],0,y.size[2])$dist

  erg <-PerformGWSinkhorn(x.distmat, y.distmat, x.weight, y.weight, epsilon, options)

  return(erg)
}
