% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{ComputeGWBarycenters}
\alias{ComputeGWBarycenters}
\title{Computing entropic Gromov-Wasserstein Barycenters}
\usage{
ComputeGWBarycenters(D, D0, lambda, epsilon, options)
}
\arguments{
\item{D:}{A List containing the distance matrices of the metric measure spaces (C_s).}

\item{D0:}{An initial distance matrix for the computations.}

\item{lambda:}{A vector containing the weights of each metric measure space in the definition of the barycenter.}

\item{epsilon:}{The regularization parameter of the entropic Gromov-Wasserstein distance, which is used to calculate the barycenter.}

\item{options:}{A named list containing several options concerning the methods of computation.
Possible options are:
\describe{
\item{"gw_loss"}{A string marking the choice of the used loss. Possibilities are the well known L2 loss "L2" (default) and the Kullback-Leibler divergence "KL".}
\item{"tol_alternating"}{A numeric value in order to stop if the progress in the alternating minimization steps is smaller than this value (default = 10^-4).}
\item{"niter_alternating"}{The max. number of steps in the alternating minimization (default = 15).}
\item{"niter"}{The max. number of projected gradient descent steps (default = 100).}
\item{"niter_sinkhorn"}{The max. number of steps in order to calculate a projection (default =400).}
\item{"tol_gw"}{A numeric value in order to stop if the progress in the minimization steps is smaller than this value (default = 10^-6).}
\item{"safe"}{A binary option in order to increase the stability of the calculations (default =0).}
\item{"gamma_init"}{A possible start coupling (as matrix) in order to speed up the calculations.}
\item{"tol_sinkhorn"}{A numeric value in order to stop if the progress in the calculation of a projection steps is smaller than this value  (defalut =10^-10).}
\item{"progress"}{A binary option to enable/disable a progressbar (default = 0, choose 1 to enable ).}
\item{"control"}{A binary option to control the progress of the calculations (default = 1). Setting this to 0 will speed up the calculations, but prevent them from being cancelled via the R console.}
}}
}
\value{
A List containing a vector with the values of the barycenter after each alternating minimization step ($meanvalue) and the final distance matrix fitting best to the underlying metric measure spaces ($msmatrix).

A possible application using the entropic Gromov-Wasserstein barycenters is the averaging of images.
Thereby one tries to find a structure, which is present in all images.
In order to do that, we subsample for each image the pixels of the images with mass, put uniform mass on them and build the corresponding distance
matrices. Then, we calculate the entropic Gromov-Wasserstein barycenter and find the best underlying distance matrix.
This distance matrix should describe the averaged structres of the images well.
One can visiualize the result using multi dimensional scaling.
}
\description{
The function \code{ComputeGWBarycenters} computes the
the entropic Gromov-Wasserstein barycenter of several discrete metric measure spaces C_s.
(cf. Algorithm 1 in “Gromov-Wasserstein Averaging of Kernel and Distance Matrices.” by Peyré,
Gabriel, Marco Cuturi, and Justin Solomon In ICML 2016, 2016.).
}
\examples{
# Computation of a entropic Gromov-Wasserstein barycenter
# based on ten images of the MNIST data set representing
#the digit 4 (included in the package)
\dontrun{n = 10
# Creating distance matricies of the subsampled images
D = sapply(MNIST_four[1:n], GWBarycenterSample, nop = 28, sample.size = 500, simplify = FALSE)

# Choosing an initial matrix and setting options
D0 = as.matrix(D[[1]])
lambda = rep(1/n,n)
epsilon = 10^(-3)
options=list("niter"=40,"niter_sinkhorn"= 100, tol_gw = 1e-4, "tol_sinkhorn"= 1e-6, "niter_alternating"=10,"tol_alternating"=1.0000e-04)

# Calling the function
A<-ComputeGWBarycenters(D, D0,lambda, epsilon,options)

# Using multidimensinal scaling in order to convert the found,
# in a sense optimal, distance matrix into an image
d <- A$msmatrix
fit <- cmdscale(d,eig=TRUE, k=2)

# Plotting the solution
x <- fit$points[,1]
y <- fit$points[,2]
plot(x,y, main="Metric MDS", ylim = c(min(x),max(x)),pch = 15, cex =5, xlab ="", ylab ="" )
}
}
\references{
“Gromov-Wasserstein Averaging of Kernel and Distance Matrices.” by Peyré,
Gabriel, Marco Cuturi, and Justin Solomon In ICML 2016, 2016.
}
