#include <RcppArmadillo.h>
#include <stdexcept>
#include "getoptions.h"
#include "barycentersupp.h"
#include "perform_gw_sinkhorn.h"
#include  "supportfct.h"
#include  "perform_gw_opttrans.h"
#define   DEBUG  0
#include <omp.h>
// [[Rcpp::plugins(openmp)]]
//[[Rcpp::depends(RcppArmadillo)]]

using namespace Rcpp;
//' @title Computing entropic Gromov-Wasserstein Barycenters
//' @description The function \code{ComputeGWBarycenters} computes the
//' the entropic Gromov-Wasserstein barycenter of several discrete metric measure spaces C_s.
//' (cf. Algorithm 1 in “Gromov-Wasserstein Averaging of Kernel and Distance Matrices.” by Peyré,
//'Gabriel, Marco Cuturi, and Justin Solomon In ICML 2016, 2016.).
//'@param D: A List containing the distance matrices of the metric measure spaces (C_s).
//'@param D0: An initial distance matrix for the computations.
//'@param lambda: A vector containing the weights of each metric measure space in the definition of the barycenter.
//'@param epsilon: The regularization parameter of the entropic Gromov-Wasserstein distance, which is used to calculate the barycenter.
//'@param options: A named list containing several options concerning the methods of computation.
//'Possible options are:
//'\describe{
//' \item{"gw_loss"}{A string marking the choice of the used loss. Possibilities are the well known L2 loss "L2" (default) and the Kullback-Leibler divergence "KL".}
//' \item{"tol_alternating"}{A numeric value in order to stop if the progress in the alternating minimization steps is smaller than this value (default = 10^-4).}
//' \item{"niter_alternating"}{The max. number of steps in the alternating minimization (default = 15).}
//' \item{"niter"}{The max. number of projected gradient descent steps (default = 100).}
//' \item{"niter_sinkhorn"}{The max. number of steps in order to calculate a projection (default =400).}
//' \item{"tol_gw"}{A numeric value in order to stop if the progress in the minimization steps is smaller than this value (default = 10^-6).}
//' \item{"safe"}{A binary option in order to increase the stability of the calculations (default =0).}
//' \item{"gamma_init"}{A possible start coupling (as matrix) in order to speed up the calculations.}
//' \item{"tol_sinkhorn"}{A numeric value in order to stop if the progress in the calculation of a projection steps is smaller than this value  (defalut =10^-10).}
//' \item{"progress"}{A binary option to enable/disable a progressbar (default = 0, choose 1 to enable ).}
//' \item{"control"}{A binary option to control the progress of the calculations (default = 1). Setting this to 0 will speed up the calculations, but prevent them from being cancelled via the R console.}
//' }
//'@return A List containing a vector with the values of the barycenter after each alternating minimization step ($meanvalue) and the final distance matrix fitting best to the underlying metric measure spaces ($msmatrix).
//'
//'A possible application using the entropic Gromov-Wasserstein barycenters is the averaging of images.
//'Thereby one tries to find a structure, which is present in all images.
//'In order to do that, we subsample for each image the pixels of the images with mass, put uniform mass on them and build the corresponding distance
//'matrices. Then, we calculate the entropic Gromov-Wasserstein barycenter and find the best underlying distance matrix.
//'This distance matrix should describe the averaged structres of the images well.
//'One can visiualize the result using multi dimensional scaling.
//'
//'
//'
//'@references “Gromov-Wasserstein Averaging of Kernel and Distance Matrices.” by Peyré,
//'Gabriel, Marco Cuturi, and Justin Solomon In ICML 2016, 2016.
//'@examples # Computation of a entropic Gromov-Wasserstein barycenter
//'# based on ten images of the MNIST data set representing
//'#the digit 4 (included in the package)
//'\dontrun{n = 10
//'# Creating distance matricies of the subsampled images
//'D = sapply(MNIST_four[1:n], GWBarycenterSample, nop = 28, sample.size = 500, simplify = FALSE)
//'
//'# Choosing an initial matrix and setting options
//'D0 = as.matrix(D[[1]])
//'lambda = rep(1/n,n)
//'epsilon = 10^(-3)
//'options=list("niter"=40,"niter_sinkhorn"= 100, tol_gw = 1e-4, "tol_sinkhorn"= 1e-6, "niter_alternating"=10,"tol_alternating"=1.0000e-04)
//'
//'# Calling the function
//'A<-ComputeGWBarycenters(D, D0,lambda, epsilon,options)
//'
//'# Using multidimensinal scaling in order to convert the found,
//'# in a sense optimal, distance matrix into an image
//'d <- A$msmatrix
//'fit <- cmdscale(d,eig=TRUE, k=2)
//'
//'# Plotting the solution
//'x <- fit$points[,1]
//'y <- fit$points[,2]
//'plot(x,y, main="Metric MDS", ylim = c(min(x),max(x)),pch = 15, cex =5, xlab ="", ylab ="" )
//'}
//[[Rcpp::export]]
List ComputeGWBarycenters(List D,arma::mat D0,arma::vec lambda,double epsilon, List options){

// compute_gw_barycenters - compute GW barycenters of distance matrices
//
//   [Dbary,R] = compute_gw_barycenters(D,lambda,epsilon,options);
//
//   R display the evolution of the minimized error.
//
//   Copyright (c) 2016 Gabriel Peyre
  std::string bary_method = getoptions_string(options, "bary_method", "alternating");
  std::string loss = getoptions_string(options, "gw_loss", "L2");

  int N0 = size(as<arma::mat>(D[0]),0);

  if(DEBUG){
    printf("D[0] is given as:\n");
    showMatrix(as<arma::mat>(D[0]));
    printf("N0 equals %i \n", N0);
  }
////
// Use uniform densities
// for barycenter
  arma::vec mu = normalize(arma::ones<arma::vec>(size(D0,0)));

  if(DEBUG){
    printf("mu equals :\n");
    showVector(mu);
  }

// for the other
  List Mu(D.size());
  Mu[0] = mu;
  if(DEBUG){
    printf("D.size equals %li\n", D.size());
    printf("M[0]equals:\n");
    showMatrix(as<arma::mat>(Mu[0]));
  }
  for (int k=0;k<(D.size());k++){
    Mu[k] = normalize(arma::ones<arma::mat>(size(as<arma::mat>(D[k]),0),1));
    if(DEBUG){
      printf("Mu[%i] is given as:\n", k);
      showMatrix(as<arma::mat>(Mu[k]));
    }
  }


/*switch bary_method
case 'bfgs'
////// L-BFGS -- depreciated ////
opts.verb = getoptions(options, 'bfgs_verb', 1);
opts.niter = getoptions(options, 'bfgs_iter', 20);
opts.bfgs_memory = getoptions(options, 'bfgs_memory', 10);
opts.report = @(f,v)v;
//
func = @(d)LossCallback_Barycenters(D,Mu,d,mu, lambda, epsilon, options);
func1 = @(x)VectorizeCallback(func,x,size(D{1}));
[d, R, info] = perform_bfgs(func1, D0(:), opts);
d = reshape(d,size(D{1}));
// set diagonal to 0
Dbary = d-diag(diag(d));
case {'alternating' 'alterating'}*/

// Preparing log domain calculation
int log_domain = getoptions_int(options, "log_domain", 0);

  if(!log_domain){
    options["epsilon"] = epsilon;
  }


////// Alternating minimization //////
  arma::mat Dbary =D0;
  double tol_alternating = getoptions_double(options, "tol_alternating", 1e-4);
  int niter_alternating = getoptions_int(options, "niter_alternating", 15);
  if(DEBUG){
    printf("tol_alternating equals %f\n", tol_alternating);
    printf("niter_alternating equals %i\n", niter_alternating);
    //printf("Dbary is initialized as: \n");
    //showMatrix(Dbary);

  }
  arma::vec R;
  R.zeros(niter_alternating);

  if(DEBUG){
    printf("R is initialized as: \n");
    showVector(R);
  }

  // Static List of length of the input
  List gamma(D.size());

  int i;
  for (i=0; i<=(niter_alternating-1); i++){
    arma::mat Dnew = zeros(size(Dbary));
    if(DEBUG){
      printf("We are in iteration i = %i \n", i);
      printf("Dnew is initialized as:\n");
      showMatrix(Dnew);
      printf("D.size is given as %li \n", D.size());
    }

    // Start of the possible parallelization
    //omp_set_num_threads(1);
    //#pragma omp parallel for shared(D,Mu,mu,epsilon,Dbary,gamma,loss,lambda,i,R,Dnew) private(options) schedule(auto) default(none)
    for (int s=0;s<D.size();s++){

      if(DEBUG){
        printf("Mu[s] is given as:\n");
        showMatrix(as<arma::mat>(Mu[s]));
      }

      // We choose a clever initialisation
      /*
      if (i==0){
        options["gamma_init"] = as<arma::mat>(Mu[s])*trans(mu);
      }
      else{
        options["gamma_init"] = as<arma::mat>(gamma[s]);
      }


      options["gamma_init"] = as<arma::mat>(Mu[s])*trans(mu);
        */
      // to avoid a lot of type casts
      arma::mat D_s = as<arma::mat>(D[s]);
      arma::vec Mu_s = as<arma::vec>(Mu[s]);

      List entr_gw;
      perform_gw_sinkhorn(D_s, Dbary, Mu_s , mu, epsilon, options, entr_gw);


      // Network simplex solver
      options["gamma_init"] = as<arma::mat>(entr_gw["gamma"]);

      if(DEBUG){
        printf("During PerformGWSinkhorn D_s is given as: \n");
        showMatrix(D_s);
        printf("During PerformGWSinkhorn Dbary is given as: \n");
        showMatrix(Dbary);
      }

      // End of construction/testing part !

      List opt_trans;
      PerformGWSinkhornOTVersion( D_s, Dbary, Mu_s , mu,  options, opt_trans);

      // Debug Stuff
      if(DEBUG){
        NumericVector tmp = opt_trans["distance"];
        printf("For %d GW is given as %.10f \n",s, tmp[tmp.size()-1] );
      }
      gamma[s]=(opt_trans["gamma"]);

      if(DEBUG){
        printf("gamma is given as: \n");
        showMatrix(opt_trans["gamma"]);
      }

      if(!loss.compare("L2")){
            //#pragma omp critical
            Dnew = Dnew + lambda(s) * trans(as<arma::mat>(gamma[s])) * (D_s) * (as<arma::mat>(gamma[s]));
      }
      else if(!loss.compare("KL")){
          //#pragma omp critical
          Dnew = Dnew + lambda(s) * trans(as<arma::mat>(gamma[s])) * log( 1e-10 + D_s ) * (as<arma::mat>(gamma[s]));
      }
       else{
          throw std::runtime_error("An unknown loss was chosen");
      }
      arma::vec dist =as<arma::vec>(opt_trans["distance"]);
      //#pragma omp critical
      R[i] = R[i] + lambda(s) *dist[dist.n_elem-1];

      if(DEBUG){
        printf("dist[dist.n_elem-1] is given as: %f\n", dist[dist.n_elem-1]);

      }
    }
    // End of the parallelisation
    Dbary = Dnew / (mu*trans(mu));
    if (!loss.compare("KL")){
      Dbary = exp(Dbary);
    }
    // early stopping
    if(i>=1)
      // printf("Abbruch kriterium for %d is given as:%f\n",i,(min(R.subvec(0,i-1)) - R[i])/std::abs(R[0]));

    if ((i>2) && (tol_alternating>0) && ((min(R.subvec(0,i-1)) - R[i])/std::abs(R[0])<tol_alternating )){
    // stop if not enough progress
      break;
    }
  }
  if(DEBUG){
    printf("The final i equals %i\n", i);
    printf("R is finally given as: \n");
    showVector(R);
  }
  R =R(arma::span(0,(i-1)));
  return(List::create(Named("msmatrix")=Dbary, Named("meanvalue")=R ));
}
