#include <RcppArmadillo.h>
#include <stdexcept>
#include "getoptions.h"
#include "supportfct.h"
#include "perform_sinkhorn.h"
#include "progressbar.cpp"
#include "perform_gw_sinkhorn.h"

#define   DEBUG  0
//[[Rcpp::depends(RcppArmadillo)]]

using namespace Rcpp;
//' @title Computing the entropic Gromov-Wasserstein distance
//' @description The function \code{PerformGWSinkhorn} computes the
//' the entropic Gromov-Wasserstein distance between two discrete metric measure spaces X and Y via
//' projected gradient descent (cf. “Gromov-Wasserstein Averaging of Kernel and Distance Matrices.” by Peyré,
//'Gabriel, Marco Cuturi, and Justin Solomon in ICML 2016, 2016.).
//'@param d1: The distance matrix of all points in X.
//'@param d2: The distance matrix of all points in Y.
//'@param mu1: The weights of the points in X as column vector.
//'@param mu2: The weights of the points in Y as column vector.
//'@param epsilon: A numeric parameter of the entropic Gromov-Wasserstein distance.
//'@param options: A named list containing several options concerning the methods of computation.
//'Possible options are:
//'\describe{
//' \item{"niter"}{The max. number of projected gradient descent steps (default = 100).}
//' \item{"niter_sinkhorn"}{The max. number of steps in order to calculate a projection (default =400).}
//' \item{"gw_loss"}{A string marking the choice of the used loss. Possibilities are the well known L2 loss "L2" (default) and the Kullback-Leibler divergence "KL".}
//' \item{"tol_gw"}{A numeric value in order to stop if the progress in the projection steps is smaller than this value (default = 10^-6).}
//' \item{"safe"}{A binary option in order to increase the stability of the calcukations (default =0).}
//' \item{"gamma_init"}{A possible start coupling (as matrix) in order to speed up the calculations.}
//' \item{"tol_sinkhorn"}{A numeric value in order to stop if the progress in the calculation of a projection steps is smaller than this value  (defalut =10^-10).}
//' \item{"progress"}{A binary option to enable/disable a progressbar (default = 0, choose 1 to enable ).}
//' \item{"control"}{A binary option to control the progress of the calculations (default = 1). Setting this to 0 will speed up the calculations, but prevent them from being cancelled via the R console.}
//' }
//'@return A List containing a vector with the values of the projection steps while calculating the entopic Gromov-Wasserstein distance  between the 2  lattices X and Y ($distance) and the final coupling ($gamma).
//'@examples
//'  We consider the well known example of the n-simplex (cf. "Gromov–Wasserstein Distances and the Metric Approach to Object Matching", by F. Mémoli in  F. Found Comput Math, 2011). This is a
//'  metric measure space consiting of n points, which have mass 1/n and distance 1 to all other points.
//'  For these sturutures some explicit expressoins and many bounds for the Gromov-Wasserstein distance are known.
//'
//'
//'  m <- 4
//'  y.distmat<- matrix(0, nrow = 1, ncol =1)
//'  x.distmat<- matrix(1, nrow = m, ncol =m)-diag(1,nrow=m, ncol =m)
//'  eps <- 0.001
//'  y.weight <- 1
//'  x.weight <- rep(1/m,m)
//'  options = list("niter"=40,"niter_sinkhorn"= 100, tol_gw = 1e-6, "tol_sinkhorn"= 1e-6, "progress" = 1)
//'
//'  PerformGWSinkhorn(x.distmat, y.distmat, x.weight, y.weight, eps, options)
//[[Rcpp::export]]
Rcpp::List PerformGWSinkhorn(arma::mat d1,arma::mat d2,arma::vec mu1,arma::vec mu2, double epsilon, List options){
  // PerformGWSinkhorn - perform iterative sinkhorn iterations to solve for GW matching
  //
  //   [gamma,a,b,GW] = PerformGWSinkhorn(d1,d2,mu1,mu2,epsilon, options);
  //
  //   Solves for (a stationary point of)
  //       GW = min_{gamma} C - <gamma*d2,d1*gamma> + epsilon*Entropy(gamma)
  //       s.t.   gamma*1=mu1 and gamma^T*1=mu2
  //   where C = 1/2*(<d1*mu1,mu1> + <d2*mu2,mu2>) is a constant offset
  //
  //   Set options.log_domain=1 to use log-domain Sinkhorn (stable even with small epsilon).
  //
  //   Copyright (c) 2015 Gabriel Peyre

  // GW = @(gamma,d1,d2)-dotp(d1*gamma,gamma*d2);

  // Check if epsilon is chosen correctly
  if (epsilon<= 0){
    throw std::runtime_error("Inadmissible choice of epsilon.");
  }
  // Choose the corresponding loss
  std::string loss = getoptions_string(options, "gw_loss", "L2");
  unsigned int chosen_loss;
  if(!loss.compare("L2")){
    chosen_loss = 0;
  }
  else{
    chosen_loss=1;
  }
  // Create an array of function pointers
  typedef arma::mat (*fptr_t)(arma::mat);
  fptr_t f_one[2] = {f_one_L2, f_one_KL};
  fptr_t f_two[2] = {f_two_L2,f_two_KL};
  fptr_t h_one[2]=  {h_one_L2, h_one_KL};
  fptr_t h_two[2] = {h_two_L2, h_two_KL};


  options["null"] = 0;
  int niter = getoptions_int(options, "niter", 100);
  int niter_sinkhorn = getoptions_int(options, "niter_sinkhorn", 400);
  double tol_gw = getoptions_double(options, "tol_gw", 1e-6);
  int safe = getoptions_int(options, "safe", 0);
  int progress = getoptions_int(options, "progress",0);
  int control = getoptions_int(options, "control", 1);

  double GWcst = dotp( f_one[chosen_loss](d1)*mu1,mu1) + dotp( f_two[chosen_loss](d2)*mu2,mu2);
  NumericVector GW;

  // initial match
  arma::mat gamma = getoptions_mat(options, "gamma_init", mu1*trans(mu2));
  // setting options for perform_sinkhorn
  double tol= getoptions_double(options, "tol_sinkhorn", 1e-10);

  if(DEBUG){
    printf("niter is set as:%i \n", niter);
    printf("niter_sinkhorn is set as:%i \n", niter_sinkhorn);
    printf("The chosen loss is:%i \n", chosen_loss);
    printf("tol_gw is set as:%f \n", tol_gw);
    printf("safe is set as:%i \n", safe);
    printf("tol_sinkhorn is set as:%f \n", tol);
    printf("safe is set as:%i \n", safe);
    printf("control is set as:%i \n", control);
    printf("gamma_init is given as:\n");
    showMatrix(gamma);
  }

  // Creating a pBar object that is used exactly when progress equals 1
  pBar bar;


  Rcpp::List opts= Rcpp::List::create(Named("niter")= niter_sinkhorn, Named("tol")=tol, Named("safe")=safe, Named("control")=control);
  Rcpp::List sinkhorn_res;

  for (int i=0; i<=niter-1; i++){

    arma::mat C = - h_one[chosen_loss](d1)*gamma*h_two[chosen_loss](trans(d2));
    arma::mat C_zero;

    C_zero = C-vectorise(C).min();
    // run sinkhorn
    sinkhorn_res = (perform_sinkhorn(C_zero,epsilon, mu1,mu2, opts));
    arma::vec a = Rcpp::as<arma::vec>((sinkhorn_res["a"]));
    arma::vec b = Rcpp::as<arma::vec>((sinkhorn_res["b"]));
    gamma = Rcpp::as<arma::mat>(sinkhorn_res["gamma"]);
    opts["a_init"] = sinkhorn_res["a"]; // warm restart

    // Checking the results of perform_sinkhorn
    if(DEBUG){
      printf("for iteration %d \n", i);
      printf("The dimension of C is %ix%i \n", size(C,0), size(C,1));
      showMatrix(C);
      printf("The interim result for gamma is:\n");
      showMatrix(gamma);

      // Checking all the results used for calculating GW
      printf("GW-interim result for Entropy(gamma):%f \n", Entropy(gamma));
      printf("GW-interim result for epsilon:%f \n", epsilon);
      printf("GW-interim result for  dotp( C,gamma ):%f \n", dotp( C,gamma ));
      printf("GW-interim result for GWcst  :%f \n", GWcst);
    }
    if(progress){
      bar.update(100.0/niter); //How much new progress was added (only needed when new progress was added)
      //Print pBar:
      bar.print(); //This should be called more frequently than it is in this demo (you'll have to see what looks best for your program)
    }
    GW.push_back(GWcst + dotp( C,gamma ) + epsilon * Entropy(gamma));


    if ((i>1) && (tol_gw>0) && ( (min(as<NumericVector>(GW[(seq_len(i))-1])) - GW(i))/(GW(1))<tol_gw )){
      // stop if not enough progress
      break;
    }
    if (isnan(gamma(0))){
      throw std::runtime_error("Computation blew up, epsilon too small");
    }
  }

  if(progress){
    bar.update((niter-GW.length())*100.0/niter); //How much new progress was added (only needed when new progress was added)
    //Print pBar:
    bar.print();
  }

  arma::mat res_gamma = (sinkhorn_res["gamma"]);
  return(List::create(Named("distance")=GW, Named("gamma")= res_gamma));
}
